﻿Imports System.Drawing.Drawing2D
Imports System.Math

Public Class Form1
    ' Display the packing.
    Private Sub Form1_Load(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles MyBase.Load
        MakeImage()
    End Sub
    Private Sub picPacking_SizeChanged(ByVal sender As Object, ByVal e As System.EventArgs) Handles picPacking.SizeChanged
        MakeImage()
    End Sub
    Private Sub MakeImage()
        Dim width As Integer = Math.Min( _
            picPacking.ClientSize.Width, _
            picPacking.ClientSize.Height)
        picPacking.Image = FindApollonianPacking(width)
    End Sub

    ' Find the Apollonian packing.
    Private Function FindApollonianPacking(ByVal width As Integer) As Bitmap
        Dim bm As New Bitmap(width, width)
        Using gr As Graphics = Graphics.FromImage(bm)
            gr.SmoothingMode = SmoothingMode.AntiAlias
            gr.Clear(Color.LightGreen)

            ' Create the three central tangent circles.
            Dim radius As Single = width * 0.225F
            Dim x As Single = width \ 2
            Dim gasket_height As Single = 2 * CSng(radius + 2 * radius / Math.Sqrt(3))
            Dim y As Single = (width - gasket_height) / 2 + radius
            Dim circle0 As New Circle(x, y, radius)

            ' Draw a box around the gasket. (For debugging.)
            'gr.DrawRectangle(Pens.Orange, _
            '    x - gasket_height / 2, _
            '    y - radius, _
            '    gasket_height, _
            '    gasket_height)

            x -= radius
            y += CSng(radius * Math.Sqrt(3))
            Dim circle1 As New Circle(x, y, radius)
            x += 2 * radius
            Dim circle2 As New Circle(x, y, radius)

            ' Find the circle that contains them all.
            Dim big_circle As Circle = FindApollonianCircle( _
                circle0, circle1, circle2, -1, -1, -1)
            big_circle.Draw(gr, Pens.Blue)

            ' Draw the big circle.
            Using the_brush As New SolidBrush(RandomColor())
                big_circle.Draw(gr, the_brush)
            End Using

            ' Draw the three central circles.
            Using the_brush As New SolidBrush(RandomColor())
                circle0.Draw(gr, the_brush)
            End Using
            Using the_brush As New SolidBrush(RandomColor())
                circle1.Draw(gr, the_brush)
            End Using
            Using the_brush As New SolidBrush(RandomColor())
                circle2.Draw(gr, the_brush)
            End Using

            ' Set level to smaller values such as 3 to see partially drawn gaskets.
            Dim level As Integer = 10000

            ' Find the central circle.
            FindCircleOutsideAll(level, gr, circle0, circle1, circle2)

            ' Find circles tangent to the big circle.
            FindCircleOutsideTwo(level, gr, circle0, circle1, big_circle)
            FindCircleOutsideTwo(level, gr, circle1, circle2, big_circle)
            FindCircleOutsideTwo(level, gr, circle2, circle0, big_circle)
        End Using
        Return bm
    End Function

    ' Draw a circle tangent to these three circles and that is outside all three.
    Private Sub FindCircleOutsideAll(ByVal level As Integer, ByVal gr As Graphics, ByVal circle0 As Circle, ByVal circle1 As Circle, ByVal circle2 As Circle)
        Dim new_circle As Circle = FindApollonianCircle(circle0, circle1, circle2, 1, 1, 1)
        If (new_circle Is Nothing) Then Return
        If (new_circle.Radius < 0.1) Then Return

        Using the_brush As New SolidBrush(RandomColor())
            new_circle.Draw(gr, the_brush)
        End Using

        level -= 1
        If (level > 0) Then
            FindCircleOutsideAll(level, gr, circle0, circle1, new_circle)
            FindCircleOutsideAll(level, gr, circle0, circle2, new_circle)
            FindCircleOutsideAll(level, gr, circle1, circle2, new_circle)
        End If
    End Sub

    ' Draw a circle tangent to these three circles and that is outside two of them.
    Private Sub FindCircleOutsideTwo(ByVal level As Integer, ByVal gr As Graphics, ByVal circle0 As Circle, ByVal circle1 As Circle, ByVal circle_contains As Circle)
        Dim new_circle As Circle = FindApollonianCircle( _
            circle0, circle1, circle_contains, 1, 1, -1)
        If (new_circle Is Nothing) Then Return
        If (new_circle.Radius < 0.1) Then Return

        Using the_brush As New SolidBrush(RandomColor())
            new_circle.Draw(gr, the_brush)
        End Using

        level -= 1
        If (level > 0) Then
            FindCircleOutsideTwo(level, gr, new_circle, circle0, circle_contains)
            FindCircleOutsideTwo(level, gr, new_circle, circle1, circle_contains)
            FindCircleOutsideAll(level, gr, circle0, circle1, new_circle)
        End If
    End Sub

    ' Find a solution to Apollonius' problem.
    ' For discussion and method, see:
    '    http://mathworld.wolfram.com/ApolloniusProblem.html
    '    http://en.wikipedia.org/wiki/Problem_of_Apollonius#Algebraic_solutions
    ' For most of a Java code implementation, see:
    '    http://www.diku.dk/hjemmesider/ansatte/rfonseca/implementations/apollonius.html
    Private Function FindApollonianCircle(ByVal c1 As Circle, ByVal c2 As Circle, ByVal c3 As Circle, ByVal s1 As Integer, ByVal s2 As Integer, ByVal s3 As Integer) As Circle
        ' Make sure c2 doesn't have the same X or Y coordinate as the others.
        Const tiny As Single = 0.0001F
        If ((Abs(c2.Center.X - c1.Center.X) < tiny) OrElse _
            (Abs(c2.Center.Y - c1.Center.Y) < tiny)) _
        Then
            Dim temp_circle As Circle = c2
            c2 = c3
            c3 = temp_circle
            Dim temp_s As Integer = s2
            s2 = s3
            s3 = temp_s
        End If
        If ((Abs(c2.Center.X - c3.Center.X) < tiny) OrElse _
            (Abs(c2.Center.Y - c3.Center.Y) < tiny)) _
        Then
            Dim temp_circle As Circle = c2
            c2 = c1
            c1 = temp_circle
            Dim temp_s As Integer = s2
            s2 = s1
            s1 = temp_s
        End If
        Debug.Assert( _
            (c2.Center.X <> c1.Center.X) AndAlso (c2.Center.Y <> c1.Center.Y) AndAlso _
            (c2.Center.X <> c3.Center.X) AndAlso (c2.Center.Y <> c3.Center.Y), _
            "Cannot find points without matching coordinates.")

        Dim x1 As Single = c1.Center.X
        Dim y1 As Single = c1.Center.Y
        Dim r1 As Single = c1.Radius
        Dim x2 As Single = c2.Center.X
        Dim y2 As Single = c2.Center.Y
        Dim r2 As Single = c2.Radius
        Dim x3 As Single = c3.Center.X
        Dim y3 As Single = c3.Center.Y
        Dim r3 As Single = c3.Radius

        Dim v11 As Single = 2 * x2 - 2 * x1
        Dim v12 As Single = 2 * y2 - 2 * y1
        Dim v13 As Single = x1 * x1 - x2 * x2 + y1 * y1 - y2 * y2 - r1 * r1 + r2 * r2
        Dim v14 As Single = 2 * s2 * r2 - 2 * s1 * r1

        Dim v21 As Single = 2 * x3 - 2 * x2
        Dim v22 As Single = 2 * y3 - 2 * y2
        Dim v23 As Single = x2 * x2 - x3 * x3 + y2 * y2 - y3 * y3 - r2 * r2 + r3 * r3
        Dim v24 As Single = 2 * s3 * r3 - 2 * s2 * r2

        Dim w12 As Single = v12 / v11
        Dim w13 As Single = v13 / v11
        Dim w14 As Single = v14 / v11

        Dim w22 As Single = v22 / v21 - w12
        Dim w23 As Single = v23 / v21 - w13
        Dim w24 As Single = v24 / v21 - w14

        Dim P As Single = -w23 / w22
        Dim Q As Single = w24 / w22
        Dim M As Single = -w12 * P - w13
        Dim N As Single = w14 - w12 * Q

        Dim a As Single = N * N + Q * Q - 1
        Dim b As Single = 2 * M * N - 2 * N * x1 + 2 * P * Q - 2 * Q * y1 + 2 * s1 * r1
        Dim c As Single = x1 * x1 + M * M - 2 * M * x1 + P * P + y1 * y1 - 2 * P * y1 - r1 * r1

        ' Find roots of a quadratic equation
        Dim solutions() As Double = QuadraticSolutions(a, b, c)
        If (solutions.Length < 1) Then Return Nothing
        Dim rs As Single = CSng(solutions(0))
        Dim xs As Single = M + N * rs
        Dim ys As Single = P + Q * rs

        Return New Circle(xs, ys, rs)
    End Function

    ' Return solutions to a quadratic equation.
    Private Function QuadraticSolutions(ByVal a As Double, ByVal b As Double, ByVal c As Double) As Double()
        Const tiny As Double = 0.000001
        Dim discriminant As Double = b * b - 4 * a * c

        ' See if there are no real solutions.
        If (discriminant < 0) Then Return New Double() {}

        ' See if there is one solution.
        If (discriminant < tiny) Then Return New Double() {-b / (2 * a)}

        ' There are two solutions.
        Return New Double() _
            { _
                (-b + Math.Sqrt(discriminant)) / (2 * a), _
                (-b - Math.Sqrt(discriminant)) / (2 * a) _
            }
    End Function

    ' Return a random color.
    Private rand As New Random()
    Private Colors() As Color = _
    { _
        Color.Red, Color.Green, Color.Blue, Color.Lime, _
        Color.Orange, Color.Fuchsia, Color.Yellow, Color.LightGreen, _
        Color.LightBlue, Color.Cyan _
    }
    Private Function RandomColor() As Color
        Return Colors(rand.Next(0, Colors.Length))
    End Function
End Class
